#ifndef yeecudah
#define yeecudah

#include <pthread.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>


#ifndef _WIN32
#include <sys/time.h>
#include <unistd.h>
#else
#include <cuda_runtime.h>
#include <cuda.h>
#endif

/* Block Configuration for CUDA */
#define BLOCKX 16
#define BLOCKY 16 


/* Use floating point precision */
#define use_float

#ifdef use_float
typedef float my_float ;
#else
typedef double my_float ;
#endif

#define Ex(I,J,K) Ex1D[ioff_e*(I) + (nz+1)*(J) + (K)]
#define Ey(I,J,K) Ey1D[ioff_e*(I) + (nz+1)*(J) + (K)]
#define Ez(I,J,K) Ez1D[ioff_e*(I) + (nz+1)*(J) + (K)]
#define Hx(I,J,K) Hx1D[ioff_h*(I) + nz*(J) + (K)]
#define Hy(I,J,K) Hy1D[ioff_h*(I) + nz*(J) + (K)]
#define Hz(I,J,K) Hz1D[ioff_h*(I) + nz*(J) + (K)]

#define cmEx(I,J,K) cEx[ioff_e*(I) + (nz+1)*(J) + (K)]
#define cmEy(I,J,K) cEy[ioff_e*(I) + (nz+1)*(J) + (K)]
#define cmEz(I,J,K) cEz[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]
#define taskEz(I,J,K) task->ez[(task->ny+1)*(task->nz+1)*(I) + (task->nz+1)*(J) + (K)]
#define cmHx(I,J,K) cHx[ioff_h*(I) + nz*(J) + (K)]
#define cmHy(I,J,K) cHy[ioff_h*(I) + nz*(J) + (K)]
#define cmHz(I,J,K) cHz[ioff_h*(I) + nz*(J) + (K)]

#define Ez1(I,J,K) tasks[0].ez[ioff_e*(I) + (nz+1)*(J) + (K)]
#define Ez2(I,J,K) tasks[1].ez[ioff_e*(I) + (nz+1)*(J) + (K)]
#define debugout(I,J,K) cEz[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]

#define ExC(I,J,K) cEx[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]
#define EyC(I,J,K) cEy[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]
#define EzC(I,J,K) cEz[(ny+1)*(nz+1)*(I) + (nz+1)*(J) + (K)]
#define HxC(I,J,K) cHx[ny*nz*(I) + nz*(J) + (K)]
#define HyC(I,J,K) cHy[ny*nz*(I) + nz*(J) + (K)]
#define HzC(I,J,K) cHz[ny*nz*(I) + nz*(J) + (K)]

#define eps0 (my_float)8.8541878E-12  /* permittivity in vacuum */
#define mu0 (my_float)1.256637061E-6  /* permeability in vacuum */
#define c0 (my_float)2.99792458E+8    /* speed of light in vacuum */
#define Const (my_float)1.0e-10
#define Tk (my_float)2.0e-9

/* Macros for accessing assignment scheduler data */
/* Number of tasks assigned */
#define TaskNum(I) workerassignment[nthreads * MAX_TASKS_PER_THREAD + nthreads - I]
/* Id of task J in current assignment for workerthread I */
#define TaskId(I,J) workerassignment[I * MAX_TASKS_PER_THREAD + J]

/* enum for timer array in task struct below */
#define TIMERS 5
enum { SETUP=0, COMPUTE=1, COPY=2, TOTAL=3, END=4 };

/* Fields in CSV file for benchmarks (does not apply to resource plot)*/
#define CSVFIELDS 3
enum { CSVN=0, CSVMFLOPS=1, CSVERR=2 };

/* Enum for run type */
enum { GREEDY=0, ASSIGNMENT=1, PLOT=2, HPLOT=3, RESOURCE=4};

/* Contains data for a task */
typedef struct {
	int nts;
	int nx, ny, nz;
	int dx, dy, dz;

	my_float *ex, *ey, *ez;
	my_float *hx, *hy, *hz;

	int bytes;
	double flop;
	double mflops;
	double t[TIMERS];

	int id;
	int workerid;
	int done;
	int calcmet_used;
} task_data;

/* Struct for containing information when calculating
 * error estimation for some benchmarks
 */
typedef struct {
	double end_time;
	double tot_time;
	int calcmet;
	double flop;
} enddata;

/* Read in main file, needed in jobs.cu, Number of OpenMP threads to be used */
extern int nomp;

/* Run the benchmark with verbose output */
//#define verbose

/* Run thourough check on GPU output, requires that output from gpu and cpu are in tasks 0 and 1 */
//#define gpucheck

int init_task(task_data *task);
int destroy_task_arrays(task_data *task);

double msecond();
double get_flop(task_data task);

void launch_threads(pthread_t *pthreads);
void join_threads(pthread_t *pthreads);

void assignment_scheduler(	int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz);
void greedy_scheduler(		int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz);
void plot_single_device(	int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz);
void plot_hybrid_setup(		int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz);
void resource_bench(		int nts, int nx, int ny, int nz, my_float dx, my_float dy, my_float dz);

void assign_workload(int i);
void printresults(int nts, double tot_time);

void *gpu_thread(void *arg);
void *cpu_thread(void *arg);
void run_cpu_job(char *output, int threadid, task_data *task);


void StartCounter();
double GetCounter();
int cmp_enddata(const void *a,const void *b);


void updateH_homo(int calcmet, int nx, int ny, int nz, int ioff_e, int ioff_h,
	my_float *Hx1D, my_float *Hy1D, my_float *Hz1D,
	my_float *Ex1D, my_float *Ey1D, my_float *Ez1D,
	my_float Cbdx, my_float Cbdy, my_float Cbdz);

void updateE_homo(int calcmet, int nx, int ny, int nz, int ioff_e, int ioff_h,
	my_float *Hx1D, my_float *Hy1D, my_float *Hz1D,
	my_float *Ex1D, my_float *Ey1D, my_float *Ez1D,
	my_float Dbdx, my_float Dbdy, my_float Dbdz);

/* Since there is only one header file
 * and the compilation uses different compilers
 * we specify which files need the CUDA entries in main.h
 */
#ifdef CUDAFILE
void CUDA_Success(cudaError_t error, char *str);
void CUDA_Success(cudaError_t error);

void run_gpu_job(char *output, int threadid, task_data *task, cudaStream_t stream);

__global__ void tid(my_float *H, int nx, int ny, int nz, my_float *np);
__global__ void updateH_cuda( my_float *cHx,  my_float *cHy,  my_float *cHz,  my_float *cEx,  my_float *cEy,  my_float *cEz, int nx, int ny, int nz, my_float Cbdx, my_float Cbdy, my_float Cbdz  );
__global__ void updateE_cuda( my_float *cHx,  my_float *cHy,  my_float *cHz,  my_float *cEx,  my_float *cEy,  my_float *cEz, int nx, int ny, int nz, my_float Dbdx, my_float Dbdy, my_float Dbdz  );

__global__ void cleanupE_yz_cuda( my_float *cHy,  my_float *cHz,  my_float *cEx, int nx, int ny, int nz, my_float Dbdy, my_float Dbdz );
__global__ void cleanupE_xz_cuda( my_float *cHx, my_float *cHz, my_float *cEy, int nx, int ny, int nz, my_float Dbdx, my_float Dbdz );
__global__ void cleanupE_xy_cuda( my_float *cHx,  my_float *cHy, my_float *cEz, int nx, int ny, int nz, my_float Dbdx, my_float Dbdy );

__global__ void apply_source( my_float *cEz, my_float p_source_factor, int ts, my_float dt, int is, int js, int ks, int ny, int nz );

#endif
#endif // yeecudah
