/*
 * CPU Kernels
 */
#include "main.h"

void updateH_homo(int calcmet, int nx, int ny, int nz, int ioff_e, int ioff_h,
                  my_float *Hx1D, my_float *Hy1D, my_float *Hz1D,
                  my_float *Ex1D, my_float *Ey1D, my_float *Ez1D,
                  my_float Cbdx, my_float Cbdy, my_float Cbdz)
{
    int i, j, k;
    /* xlC (IBM) Not recognized by SGI-cc */
#pragma disjoint(*Ex1D, *Ey1D, *Ez1D, *Hx1D, *Hy1D, *Hz1D)

    if (calcmet==4)
    {

        for(i=0; i<nx; i++)
            for(j=0; j<ny; j++)
                for(k=0; k<nz; k++)
                    Hx(i,j,k) += (Ey(i,j,k+1)-Ey(i, j ,k))*Cbdz +
                                 (Ez(i,j, k )-Ez(i,j+1,k))*Cbdy;

        for(i=0; i<nx; i++)
            for(j=0; j<ny; j++)
                for(k=0; k<nz; k++)
                    Hy(i,j,k) += (Ez(i+1,j,k)-Ez(i,j, k ))*Cbdx +
                                 (Ex( i ,j,k)-Ex(i,j,k+1))*Cbdz;

        for(i=0; i<nx; i++)
            for(j=0; j<ny; j++)
                for(k=0; k<nz; k++)
                    Hz(i,j,k) += (Ex(i,j+1,k)-Ex( i ,j,k))*Cbdy +
                                 (Ey(i, j ,k)-Ey(i+1,j,k))*Cbdx;
    }

    /* The kernel we use in our scheduler */
    else if (calcmet==5)
    {
        /* Use OpenMP on this for loop */
        #pragma omp for
        for(i=0; i<nx; i++)
              for(j=0; j<ny; j++)
                  for(k=0; k<nz; k++)
                  {
                      Hx(i,j,k) += (Ey(i,j,k+1)-Ey(i, j ,k))*Cbdz +
                                   (Ez(i,j, k )-Ez(i,j+1,k))*Cbdy;
                      Hy(i,j,k) += (Ez(i+1,j,k)-Ez(i,j, k ))*Cbdx +
                                   (Ex( i ,j,k)-Ex(i,j,k+1))*Cbdz;
                      Hz(i,j,k) += (Ex(i,j+1,k)-Ex( i ,j,k))*Cbdy +
                                   (Ey(i, j ,k)-Ey(i+1,j,k))*Cbdx;
                  }
    }
    else
    {
        printf("FATAL ERROR! Incorrect calcmet number in updateH_homo\n");
        printf("calcmet = %i\n", calcmet);
        exit(1);
    }
}

void updateE_homo(int calcmet, int nx, int ny, int nz, int ioff_e, int ioff_h,
                  my_float *Hx1D, my_float *Hy1D, my_float *Hz1D,
                  my_float *Ex1D, my_float *Ey1D, my_float *Ez1D,
                  my_float Dbdx, my_float Dbdy, my_float Dbdz)
{
    int i, j, k;
    /* xlC (IBM) Not recognized by SGI-cc */
#pragma disjoint(*Ex1D, *Ey1D, *Ez1D, *Hx1D, *Hy1D, *Hz1D)

    if (calcmet==4)
    {

        for(i=0; i<nx; i++)
            for(j=1; j<ny; j++)
                for(k=1; k<nz; k++)
                    Ex(i,j,k) += (Hz(i,j, k )-Hz(i,j-1,k))*Dbdy +
                                 (Hy(i,j,k-1)-Hy(i, j ,k))*Dbdz;

        for(i=1; i<nx; i++)
            for(j=0; j<ny; j++)
                for(k=1; k<nz; k++)
                    Ey(i,j,k) += (Hx( i ,j,k)-Hx(i,j,k-1))*Dbdz +
                                 (Hz(i-1,j,k)-Hz(i,j, k ))*Dbdx;

        for(i=1; i<nx; i++)
            for(j=1; j<ny; j++)
                for(k=0; k<nz; k++)
                    Ez(i,j,k) += (Hy(i, j ,k)-Hy(i-1,j,k))*Dbdx +
                                 (Hx(i,j-1,k)-Hx( i ,j,k))*Dbdy;
    }
    
    /* The kernel we use in our scheduler */
    else if (calcmet==5)
    {
      
      /* Use OpenMP on this for loop */
      #pragma omp for
      for(j=1; j<ny; j++)
          for(k=1; k<nz; k++)
              Ex(0,j,k) += (Hz(0,j, k )-Hz(0,j-1,k))*Dbdy +
                           (Hy(0,j,k-1)-Hy(0, j ,k))*Dbdz;
      
      /* Use OpenMP on this for loop */
      #pragma omp for
      for(i=1; i<nx; i++)
          for(k=1; k<nz; k++)
              Ey(i,0,k) += (Hx( i ,0,k)-Hx(i,0,k-1))*Dbdz +
                           (Hz(i-1,0,k)-Hz(i,0, k ))*Dbdx;
      
      /* Use OpenMP on this for loop */
      #pragma omp for
      for(i=1; i<nx; i++)
          for(j=1; j<ny; j++)
              Ez(i,j,0) += (Hy(i, j ,0)-Hy(i-1,j,0))*Dbdx +
                           (Hx(i,j-1,0)-Hx( i ,j,0))*Dbdy;
      
      /* Use OpenMP on this for loop */
      #pragma omp for
      for(i=1; i<nx; i++)
          for(j=1; j<ny; j++)
              for(k=1; k<nz; k++)
              {
                  Ex(i,j,k) += (Hz(i,j, k )-Hz(i,j-1,k))*Dbdy +
                               (Hy(i,j,k-1)-Hy(i, j ,k))*Dbdz;
                  Ey(i,j,k) += (Hx( i ,j,k)-Hx(i,j,k-1))*Dbdz +
                               (Hz(i-1,j,k)-Hz(i,j, k ))*Dbdx;
                  Ez(i,j,k) += (Hy(i, j ,k)-Hy(i-1,j,k))*Dbdx +
                               (Hx(i,j-1,k)-Hx( i ,j,k))*Dbdy;
              }
    }
    else
    {
        printf("FATAL ERROR! Incorrect calcmet number in updateE_homo\n");
        printf("calcmet = %i\n", calcmet);
        exit(1);
    }
}
