/* CUDA Kernels */
#define CUDAFILE
#include "main.h"

__global__ void updateH_cuda(my_float *cHx,  my_float *cHy,  my_float *cHz,  my_float *cEx,  my_float *cEy,  my_float *cEz, int nx, int ny, int nz, my_float Cbdx, my_float Cbdy, my_float Cbdz )
{

	int i = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + (threadIdx.x + blockDim.x * threadIdx.y);
	
	int y = i/nz;
	int z = i%nz;

	if (!(y < ny && z < nz)) return;
	
	for(int x=0; x<nx; x++)
	{
		__syncthreads();
		HxC(x,y,z) += (EyC(x,y,z+1)-EyC(x, y ,z))*Cbdz +
						(EzC(x,y, z )-EzC(x,y+1,z))*Cbdy;

		HyC(x,y,z) += (EzC(x+1,y,z)-EzC(x,y, z ))*Cbdx +
						(ExC( x ,y,z)-ExC(x,y,z+1))*Cbdz;

		HzC(x,y,z) += (ExC(x,y+1,z)-ExC( x ,y,z))*Cbdy +
						(EyC(x, y ,z)-EyC(x+1,y,z))*Cbdx;
	}
}

__global__ void updateE_cuda( my_float *cHx,  my_float *cHy,  my_float *cHz,  my_float *cEx,  my_float *cEy,  my_float *cEz, int nx, int ny, int nz,  my_float Dbdx, my_float Dbdy, my_float Dbdz)
{

	int i = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + (threadIdx.x + blockDim.x * threadIdx.y);
	
	int y = i/nz + 1;
	int z = i%nz + 1;
	
	if (!(y < ny && z < nz)) return;

	for(int x=1; x<nx; x++)
	{
		__syncthreads();
		ExC(x,y,z) += (HzC(x,y, z )-HzC(x,y-1,z))*Dbdy +
						(HyC(x,y,z-1)-HyC(x, y ,z))*Dbdz;

		EyC(x,y,z) += (HxC( x ,y,z)-HxC(x,y,z-1))*Dbdz +
						(HzC(x-1,y,z)-HzC(x,y, z ))*Dbdx;

		EzC(x,y,z) += (HyC(x, y ,z)-HyC(x-1,y,z))*Dbdx +
						(HxC(x,y-1,z)-HxC( x ,y,z))*Dbdy;

	}
}

__global__ void cleanupE_yz_cuda( my_float *cHy,  my_float *cHz,  my_float *cEx, int nx, int ny, int nz, my_float Dbdy, my_float Dbdz )
{
	
	int i = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + (threadIdx.x + blockDim.x * threadIdx.y);
	
	int a = i/nz + 1;
	int b = i%nz + 1;

	if (a < ny && b < nz)
	ExC(0,a,b) += (HzC(0,a, b )-HzC(0,a-1,b))*Dbdy +
				   (HyC(0,a,b-1)-HyC(0, a ,b))*Dbdz;
}

__global__ void cleanupE_xz_cuda( my_float *cHx, my_float *cHz, my_float *cEy, int nx, int ny, int nz,  my_float Dbdx, my_float Dbdz )
{
	
	int i = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + (threadIdx.x + blockDim.x * threadIdx.y);
	
	int a = i/nz + 1;
	int b = i%nz + 1;

	if (a < nx && b < nz)
	EyC(a,0,b) += (HxC( a ,0,b)-HxC(a,0,b-1))*Dbdz +
					(HzC(a-1,0,b)-HzC(a,0, b ))*Dbdx;
}

__global__ void cleanupE_xy_cuda( my_float *cHx,  my_float *cHy, my_float *cEz, int nx, int ny, int nz,  my_float Dbdx, my_float Dbdy )
{
	
	int i = (blockIdx.x  + gridDim.x  * blockIdx.y) * (blockDim.x * blockDim.y) + (threadIdx.x + blockDim.x * threadIdx.y);
	
	int a = i/ny + 1;
	int b = i%ny + 1;

	if (a < nx && b < ny)
	EzC(a,b,0) += (HyC(a, b ,0)-HyC(a-1,b,0))*Dbdx +
					(HxC(a,b-1,0)-HxC( a ,b,0))*Dbdy;
}

__global__ void apply_source( my_float *cEz, my_float p_source_factor, int ts, my_float dt, int is, int js, int ks, int ny, int nz )
{
    EzC(is, js, ks) += - p_source_factor *( (ts*dt-3.0*Tk)/(Tk*Tk) )
            *exp( -1.0*(ts*dt-3.0*Tk)*(ts*dt-3.0*Tk)/(Tk*Tk) );
}
